/*

 x1541.c - A mostly working PCLink server for IDEDOS 0.9x

 Written by
  Kajtar Zsolt <soci@c64.rulez.org>

    This file is part of IDEDOS the IDE64 disk operating system
    See README for copyright notice.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <errno.h>

#ifndef WIN32
#include <sys/io.h>
#else
#include <windows.h>
#endif

#include <unistd.h>

#include "x1541.h"
#include "crc8.h"

extern volatile char tout;
static unsigned short int IOPORT, STATPORT, CTRLPORT, PORTIN;
static unsigned char DATALO, DATAHI, CLKLO, CLKHI, ATNHI, DATA_, CLK_, ATN_, DATA;
#define CLK 2


#ifdef WIN32
  #ifndef WIN32_IO

/* prototype (function typedef) for DLL function Inp32: */

typedef short _stdcall (*inpfuncPtr)(short portaddr);
typedef void _stdcall (*oupfuncPtr)(short portaddr, short datum);

HINSTANCE hLib;
inpfuncPtr inp32;
oupfuncPtr oup32;

void outb(unsigned char value, int port);
unsigned char inb(int port);

/* outb, inb - already defined in pc64.c

void outb(unsigned char value, int port)
{
    (oup32)(port,value);
}

unsigned char inb(int port)
{
    return (inp32)(port);
}
*/

  #else

static __inline__ void
outb(unsigned char _data, unsigned short _port)
{
    __asm__ __volatile__ ("outb %1, %0"
  :
  : "dN" (_port),
    "a" (_data));
}

static __inline__ unsigned char
inb(unsigned short _port)
{
  unsigned char rv;
  __asm__ __volatile__ ("inb %1, %0"
	  : "=a" (rv)
	  : "dN" (_port));
  return rv;
}

  #endif // !WIN32_IO
#else // WIN32
static int parportfd = -1;
static int parportin;


#include <linux/ppdev.h>
#include <linux/parport.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#endif // WIN32

static int inv;

static int vtime=1;

static int getbparport(void);
static int sendbparport(unsigned char b);
static int getb(void);
static int sendb(unsigned char b);

static char *drivername(void) {
    return "IEC";
}

static int initialize(char *dev, int port) {

    if (port & 0x10000) vtime=0; //hog
    IOPORT = port & 0xffff;
    STATPORT = IOPORT + 1;
    CTRLPORT = IOPORT + 2;
    switch (port & 0x40000) {
    case 0x00000://X1541
        DATAHI=0;DATALO=8;DATA=8;
        CLKHI=0;CLKLO=2;
        ATNHI=0;
        DATA_=8;CLK_=2;ATN_=1;
        inv=0;PORTIN=CTRLPORT;
#ifndef WIN32
        parportin=PPRCONTROL;
#endif
        break;
    case 0x40000://XE1541
        DATAHI=0;DATALO=8;DATA=8;
        CLKHI=0;CLKLO=2;
        ATNHI=0;
        DATA_=0x80;CLK_=0x20;ATN_=0x10;
        inv=-1;PORTIN=STATPORT;
#ifndef WIN32
        parportin=PPRSTATUS;
#endif
        break;
    case 0x80000://XM1541
        DATAHI=4;DATALO=0;DATA=4;
        CLKHI=0;CLKLO=2;
        ATNHI=0;
        DATA_=0x40;CLK_=0x20;ATN_=0x10;
        inv=-1;PORTIN=STATPORT;
#ifndef WIN32
        parportin=PPRSTATUS;
#endif
        break;
    case 0xc0000://XA1541
        DATAHI=0;DATALO=4;DATA=4;
        CLKHI=2;CLKLO=0;
        ATNHI=1;
        DATA_=0x40;CLK_=0x20;ATN_=0x10;
        inv=-1;PORTIN=STATPORT;
#ifndef WIN32
        parportin=PPRSTATUS;
#endif
        break;
    }
  #ifdef WIN32
   #ifndef WIN32_IO
    /* Load the library */
    hLib = LoadLibrary("inpout32.dll");
    if (hLib == NULL) {fprintf(stderr,"LoadLibrary Failed.\n");exit(1);}

    /* get the address of the function */

    inp32 = (inpfuncPtr) GetProcAddress(hLib, "Inp32");
    if (inp32 == NULL) {fprintf(stderr,"GetProcAddress for Inp32 Failed.\n");exit(1);}

    oup32 = (oupfuncPtr) GetProcAddress(hLib, "Out32");
    if (oup32 == NULL) {fprintf(stderr,"GetProcAddress for Oup32 Failed.\n");exit(1);}
   #endif // !WIN32_IO
  #else

    if (port & 0x20000) {
        parportfd=open(dev, O_RDWR);
        if (parportfd == -1) {
            fprintf(stderr,"Opening parport device\n");
            return -ENODEV;
        }
        if (ioctl(parportfd, PPEXCL) || ioctl(parportfd, PPCLAIM)) {
            close(parportfd);
            fprintf(stderr,"Cannot access parport device\n");
            return -ENODEV;
        }
        {
            int mode=IEEE1284_MODE_ECP;
            if (ioctl(parportfd, PPSETMODE, &mode)) {
                close(parportfd);
                fprintf(stderr,"Cannot set byte mode on parport device\n");
                return -ENODEV;
            }

            mode=0;
            if (ioctl(parportfd, PPDATADIR, &mode)) {
                close(parportfd);
                fprintf(stderr,"Cannot set direction on parport device\n");
                return -ENODEV;
            }
        }
        x1541_io_ops.getb=getbparport;
        x1541_io_ops.sendb=sendbparport;
    } else {
        parportfd = -1;
        if (ioperm(IOPORT,3,1)) {
            fprintf(stderr,"Cannot allocate IO ports\n");
            return -ENODEV;
        }
        x1541_io_ops.getb=getb;
        x1541_io_ops.sendb=sendb;
    }
  #endif // WIN32 (LPT port)    
    return 0;
}

#ifndef WIN32
static inline int watnparport(int i) {
    unsigned char b;
    while (!tout) {
        ioctl(parportfd, parportin, &b);
        if (((b ^ inv) & ATN_)!=i) return 0;
    }
    return -EIO;
}

static int getbparport() {
    unsigned char b=0, dl=DATALO+CLKHI+ATNHI, dh=DATAHI+CLKHI+ATNHI, c;
    ioctl(parportfd, PPWCONTROL, &dl);
    alarm(1);

    if (watnparport(0)) return -EIO;
    ioctl(parportfd, parportin, &c);
    if (c & CLK_) b|=128;
    ioctl(parportfd, PPWCONTROL, &dh);

    if (watnparport(ATN_)) return -EIO;
    ioctl(parportfd, parportin, &c);
    if (c & CLK_) b|=64;
    ioctl(parportfd, PPWCONTROL, &dl);
    if ((c ^ inv) & DATA_) tout=1;

    if (watnparport(0)) return -EIO;
    ioctl(parportfd, parportin, &c);
    if (c & CLK_) b|=32;
    ioctl(parportfd, PPWCONTROL, &dh);

    if (watnparport(ATN_)) return -EIO;
    ioctl(parportfd, parportin, &c);
    if (c & CLK_) b|=16;
    ioctl(parportfd, PPWCONTROL, &dl);
    if ((c ^ inv) & DATA_) tout=1;

    if (watnparport(0)) return -EIO;
    ioctl(parportfd, parportin, &c);
    if (c & CLK_) b|=8;
    ioctl(parportfd, PPWCONTROL, &dh);

    if (watnparport(ATN_)) return -EIO;
    ioctl(parportfd, parportin, &c);
    if (c & CLK_) b|=4;
    ioctl(parportfd, PPWCONTROL, &dl);
    if ((c ^ inv) & DATA_) tout=1;

    if (watnparport(0)) return -EIO;
    ioctl(parportfd, parportin, &c);
    if (c & CLK_) b|=2;
    ioctl(parportfd, PPWCONTROL, &dh);

    if (watnparport(ATN_)) return -EIO;
    ioctl(parportfd, parportin, &c);
    if (c & CLK_) b|=1;
    if ((c ^ inv) & DATA_) tout=1;
    b^=inv;
    crc_add(b);
    return b;
}

static int sendbparport(unsigned char a) {
    int b, c;
    b=a ^ (a << 1);
    crc_add(a);
    c=a & 128?(CLKHI | ATNHI | DATAHI):(CLKLO | ATNHI | DATAHI);
    ioctl(parportfd, PPWCONTROL, &c);c^=DATA;ioctl(parportfd, PPWCONTROL, &c);
    alarm(1);
    if (b & 128) c^=CLK;
    if (watnparport(0)) return -EIO;
    if (b & 128) ioctl(parportfd, PPWCONTROL, &c);c^=DATA;ioctl(parportfd, PPWCONTROL, &c);
//    if (in() & DATA_) tout=1;
    if (b & 64) c^=CLK;
    if (watnparport(ATN_)) return -EIO;
    if (b & 64) ioctl(parportfd, PPWCONTROL, &c);c^=DATA;ioctl(parportfd, PPWCONTROL, &c);
    if (b & 32) c^=CLK;
    if (watnparport(0)) return -EIO;
    if (b & 32) ioctl(parportfd, PPWCONTROL, &c);c^=DATA;ioctl(parportfd, PPWCONTROL, &c);
    if (b & 16) c^=CLK;
//    if (in() & DATA_) tout=1;
    if (watnparport(ATN_)) return -EIO;
    if (b & 16) ioctl(parportfd, PPWCONTROL, &c);c^=DATA;ioctl(parportfd, PPWCONTROL, &c);
    if (b & 8) c^=CLK;
    if (watnparport(0)) return -EIO;
    if (b & 8) ioctl(parportfd, PPWCONTROL, &c);c^=DATA;ioctl(parportfd, PPWCONTROL, &c);
//    if (in() & DATA_) tout=1;
    if (b & 4) c^=CLK;
    if (watnparport(ATN_)) return -EIO;
    if (b & 4) ioctl(parportfd, PPWCONTROL, &c);c^=DATA;ioctl(parportfd, PPWCONTROL, &c);
    if (b & 2) c^=CLK;
    if (watnparport(0)) return -EIO;
    if (b & 2) ioctl(parportfd, PPWCONTROL, &c);c^=DATA;ioctl(parportfd, PPWCONTROL, &c);
//    if (in() & DATA_) tout=1;
    if (watnparport(ATN_)) return -EIO;
    return 0;
}
#endif

static inline int watn(int i) {
    unsigned char b;
    while (!tout) {
        b=inb(PORTIN);
        if (((b ^ inv) & ATN_)!=i) return 0;
    }
    return -EIO;
}

static int getb() {
    unsigned char b=0,c;
    outb(DATALO+CLKHI+ATNHI, CTRLPORT);
    alarm(1);

    if (watn(0)) return -EIO;
    if (inb(PORTIN) & CLK_) b|=128;
    outb(DATAHI+CLKHI+ATNHI, CTRLPORT);

    if (watn(ATN_)) return -EIO;
    c=inb(PORTIN);
    if (c & CLK_) b|=64;
    outb(DATALO+CLKHI+ATNHI, CTRLPORT);
    if ((c ^ inv) & DATA_) tout=1;

    if (watn(0)) return -EIO;
    if (inb(PORTIN) & CLK_) b|=32;
    outb(DATAHI+CLKHI+ATNHI, CTRLPORT);

    if (watn(ATN_)) return -EIO;
    c=inb(PORTIN);
    if (c & CLK_) b|=16;
    outb(DATALO+CLKHI+ATNHI, CTRLPORT);
    if ((c ^ inv) & DATA_) tout=1;

    if (watn(0)) return -EIO;
    if (inb(PORTIN) & CLK_) b|=8;
    outb(DATAHI+CLKHI+ATNHI, CTRLPORT);

    if (watn(ATN_)) return -EIO;
    c=inb(PORTIN);
    if (c & CLK_) b|=4;
    outb(DATALO+CLKHI+ATNHI, CTRLPORT);
    if ((c ^ inv) & DATA_) tout=1;

    if (watn(0)) return -EIO;
    if (inb(PORTIN) & CLK_) b|=2;
    outb(DATAHI+CLKHI+ATNHI, CTRLPORT);

    if (watn(ATN_)) return -EIO;
    c=inb(PORTIN);
    if (c & CLK_) b|=1;
    if ((c ^ inv) & DATA_) tout=1;
    b^=inv;
    crc_add(b);
    return b;
}

static int sendb(unsigned char a) {
    int b, c;
    b=a ^ (a << 1);
    crc_add(a);
    c=a & 128?(CLKHI | ATNHI | DATALO):(CLKLO | ATNHI | DATALO);
    outb(c ^ DATA, CTRLPORT);outb(c, CTRLPORT);
    alarm(1);
    if (b & 128) c^=CLK;
    if (watn(0)) return -EIO;
    if (b & 128) outb(c, CTRLPORT);outb(c ^ DATA, CTRLPORT);
//    if (in() & DATA_) tout=1;
    if (b & 64) c^=CLK;
    if (watn(ATN_)) return -EIO;
    if (b & 64) outb(c ^ DATA, CTRLPORT);outb(c, CTRLPORT);
    if (b & 32) c^=CLK;
    if (watn(0)) return -EIO;
    if (b & 32) outb(c, CTRLPORT);outb(c ^ DATA, CTRLPORT);
    if (b & 16) c^=CLK;
//    if (in() & DATA_) tout=1;
    if (watn(ATN_)) return -EIO;
    if (b & 16) outb(c ^ DATA, CTRLPORT);outb(c, CTRLPORT);
    if (b & 8) c^=CLK;
    if (watn(0)) return -EIO;
    if (b & 8) outb(c, CTRLPORT);outb(c ^ DATA, CTRLPORT);
//    if (in() & DATA_) tout=1;
    if (b & 4) c^=CLK;
    if (watn(ATN_)) return -EIO;
    if (b & 4) outb(c ^ DATA, CTRLPORT);outb(c, CTRLPORT);
    if (b & 2) c^=CLK;
    if (watn(0)) return -EIO;
    if (b & 2) outb(c, CTRLPORT);outb(c ^ DATA, CTRLPORT);
//    if (in() & DATA_) tout=1;
    if (watn(ATN_)) return -EIO;
    return 0;
}

static int getbytes(unsigned short int bytes, unsigned char *data) {
    int i, j;
    for (i=0;i<bytes;i++) {
        j=x1541_io_ops.getb();
        if (j<0) return j;
        data[i]=j;
    }
    return 0;
}

static int sendbytes(unsigned short int bytes, unsigned char *data) {
    int i, j;
    for (i=0;i<bytes;i++) {
        j=x1541_io_ops.sendb(data[i]);
        if (j) return j;
    }
    return 0;
}

static void shutdown(void) {
#ifndef WIN32
    if (parportfd >= 0) {
        ioctl(parportfd, PPRELEASE);
        close(parportfd);
    }
#endif
    return;
}

static int flush(void) {
    alarm(0);
    return 0;
}

static void clean(void) {
    return;
}

static int wait(void) {
    int dyntime=1;
    int i;
start:
    alarm(0);
#ifndef WIN32
    if (parportfd >= 0) {
        unsigned char c;
        c=CLKLO+DATAHI+ATNHI;ioctl(parportfd, PPWCONTROL, &c);
        for (;;) {
            ioctl(parportfd, parportin, &c);
            if ((c ^ inv) & DATA_) break;
            if (vtime) {
                usleep(dyntime);
                if (dyntime<16384) dyntime<<=1;
            }
        }
        c=CLKHI+DATAHI+ATNHI;ioctl(parportfd, PPWCONTROL, &c);
        tout=0;alarm(1);
        for (;;) {
            ioctl(parportfd, parportin, &c);
            if (((c ^ inv) & DATA_)==0) break;
            if (vtime) {
                usleep(dyntime);
                if (dyntime<16384) dyntime<<=1;
            }
            if (tout) goto start;
        } //set clklo, wait atnlo
    } else
#endif
    {
        outb(CLKLO+DATAHI+ATNHI, CTRLPORT);
        while (((inb(PORTIN) ^ inv) & DATA_)==0) if (vtime) {
            usleep(dyntime);
            if (dyntime<16384) dyntime<<=1;
        }
        outb(CLKHI+DATAHI+ATNHI, CTRLPORT);
        tout=0;alarm(1);
        while (((inb(PORTIN) ^ inv) & DATA_)!=0) {
            if (vtime) {
                usleep(dyntime);
                if (dyntime<16384) dyntime<<=1;
            }
            if (tout) goto start;
        } //set clklo, wait atnlo
    }
    i=x1541_io_ops.getb();
    if (tout) goto start;
    alarm(0);
    return (unsigned char)i;
}

struct io_ops x1541_io_ops = {
    .drivername = drivername,
    .initialize = initialize,
    .getb       = getb,
    .getbytes   = getbytes,
    .sendb      = sendb,
    .sendbytes  = sendbytes,
    .shutdown   = shutdown,
    .flush      = flush,
    .wait       = wait,
    .clean      = clean,
};
